from transformer.plugins import plugin, Contract
from transformer.request import Header
from transformer.task import Task2


@plugin(Contract.OnTask)
def plugin(task: Task2) -> Task2:
    """
    Removes Chrome-specific, RFC-non-compliant headers starting with `:`.
    Converts header names to lowercase to simplify further overriding.
    Removes the cookie header as it is handled by Locust's HttpSession.
    """
    task.request.headers = [
        Header(name=h.name.lower(), value=h.value)
        for h in task.request.headers
        if not h.name.startswith(":") and h.name.lower() != "cookie"
    ]

    return task



from datetime import datetime
from urllib.parse import urlparse

from transformer.request import HttpMethod, Header, Request
from transformer.task import Task2
from .sanitize_headers import plugin


def test_its_name_is_resolvable():
    from transformer.plugins import resolve

    assert list(resolve("transformer.plugins.sanitize_headers")) == [plugin]


TS = datetime(1970, 1, 1)


def task_with_header(name: str, value: str) -> Task2:
    return Task2(
        name="some task",
        request=Request(
            timestamp=TS,
            method=HttpMethod.GET,
            url=urlparse("https://example.com"),
            name="task_name",
            headers=[Header(name=name, value=value)],
            post_data={},
            query=[],
        ),
    )


def test_it_removes_headers_beginning_with_a_colon():
    task = task_with_header(":non-rfc-header", "some value")
    sanitized_headers = plugin(task).request.headers
    assert len(sanitized_headers) == 0


def test_it_downcases_header_names():
    task = task_with_header("Some Name", "some value")
    sanitized_headers = plugin(task).request.headers
    header_names = {h.name for h in sanitized_headers}
    assert "some name" in header_names


def test_it_removes_cookies():
    task = task_with_header("Cookie", "some value")
    sanitized_headers = plugin(task).request.headers
    assert len(sanitized_headers) == 0


def test_it_does_not_change_nor_remove_other_headers():
    task = task_with_header("some other header", "some value")
    sanitized_headers = plugin(task).request.headers
    assert len(sanitized_headers) == 1



import pytest
from hypothesis import given
from hypothesis.strategies import from_type

from transformer.plugins import apply, group_by_contract
from .contracts import (
    Contract,
    plugin,
    contract,
    InvalidContractError,
    InvalidPluginError,
)


@given(from_type(Contract))
def test_contract_returns_contract_associated_with_plugin_decorator(c: Contract):
    @plugin(c)
    def foo():
        ...

    assert contract(foo) is c


def test_plugin_decorator_raises_with_invalid_contract():
    with pytest.raises(InvalidContractError):

        @plugin(2)
        def foo():
            ...


def test_plugin_decorator_raises_without_contract():
    with pytest.raises(InvalidContractError):

        @plugin
        def foo():
            ...


def test_contract_raises_with_invalid_plugin():
    def foo():
        ...

    with pytest.raises(InvalidPluginError):
        contract(foo)


def test_plugin_is_exported_by_the_transformer_plugins_module():
    try:
        from transformer.plugins import plugin
    except ImportError:
        pytest.fail("plugin should be exported by transformer.plugins")


def test_Contract_is_exported_by_the_transformer_plugins_module():
    try:
        from transformer.plugins import Contract
    except ImportError:
        pytest.fail("Contract should be exported by transformer.plugins")


class TestApply:
    def test_return_init_unchanged_without_plugins(self):
        x = object()
        assert apply([], x) is x

    def test_return_plugin_result(self):
        @plugin(Contract.OnTask)
        def plugin_a(x: str) -> str:
            return x + "a"

        assert apply([plugin_a], "z") == "za"

    def test_runs_plugins_in_succession_on_input(self):
        @plugin(Contract.OnTask)
        def plugin_a(x: str) -> str:
            return x + "a"

        @plugin(Contract.OnTask)
        def plugin_b(x: str) -> str:
            return x + "b"

        assert apply((plugin_a, plugin_b), "") == "ab"
        assert apply((plugin_b, plugin_a, plugin_b), "") == "bab"


class TestGroupByContract:
    def test_return_empty_dict_when_no_plugins(self):
        assert group_by_contract([]) == {}

    def test_index_plugins_with_simple_contracts_by_their_contract(self):
        @plugin(Contract.OnTask)
        def plugin_a():
            pass

        @plugin(Contract.OnTask)
        def plugin_b():
            pass

        @plugin(Contract.OnScenario)
        def plugin_z():
            pass

        assert group_by_contract((plugin_a, plugin_b, plugin_z)) == {
            Contract.OnScenario: [plugin_z],
            Contract.OnTask: [plugin_a, plugin_b],
        }

    def test_index_plugins_with_complex_contracts_by_their_basic_contracts(self):
        @plugin(Contract.OnTask)
        def plugin_task():
            pass

        @plugin(Contract.OnTask | Contract.OnScenario | Contract.OnPythonProgram)
        def plugin_multi():
            pass

        assert group_by_contract((plugin_task, plugin_multi)) == {
            Contract.OnTask: [plugin_task, plugin_multi],
            Contract.OnScenario: [plugin_multi],
            Contract.OnPythonProgram: [plugin_multi],
        }



"""
:mod:`transformer` -- Main API
============================================

This module exports the functions that should cover most use-cases of any
Transformer user.
"""
import pkg_resources
from .transform import dumps, dump

__version__ = pkg_resources.get_distribution("har-transformer").version

__all__ = ["dumps", "dump"]



# pylint: skip-file
import enum
import io
from unittest.mock import MagicMock, Mock
from unittest.mock import patch
from urllib.parse import urlparse

import pytest
from hypothesis import given
from hypothesis.strategies import composite, sampled_from, booleans

from transformer import python as py
from transformer.request import Header, QueryPair
from transformer.task import (
    Task,
    Request,
    HttpMethod,
    TIMEOUT,
    LocustRequest,
    Task2,
    RequestsPostData,
    JSON_MIME_TYPE,
    req_to_expr,
    lreq_to_expr,
)


class TestTask:
    class TestFromRequests:
        def test_it_returns_a_task(self):
            request = MagicMock()
            request.timestamp = 1
            second_request = MagicMock()
            second_request.timestamp = 2
            assert all(
                isinstance(t, Task)
                for t in Task.from_requests([request, second_request])
            )

        @patch("builtins.open")
        def test_it_doesnt_create_a_task_if_the_url_is_on_the_blacklist(
            self, mock_open
        ):
            mock_open.return_value = io.StringIO("amazon")
            request = MagicMock()
            request.url = MagicMock()
            request.url.netloc = "www.amazon.com"
            task = Task.from_requests([request])
            assert len(list(task)) == 0

        @patch("builtins.open")
        def test_it_creates_a_task_if_the_path_not_host_is_on_the_blacklist(
            self, mock_open
        ):
            mock_open.return_value = io.StringIO("search\namazon")
            request = MagicMock()
            request.url = urlparse("https://www.google.com/search?&q=amazon")
            task = Task.from_requests([request])
            assert len(list(task)) == 1

    class TestReplaceURL:
        def test_it_creates_a_locust_request_when_there_is_none(self):
            task = Task(name="some name", request=MagicMock())

            modified_task = Task.replace_url(task, "")

            assert modified_task.locust_request

        def test_it_returns_a_task_with_the_replaced_url(self):
            locust_request = LocustRequest(
                method=MagicMock(), url=MagicMock(), headers=MagicMock()
            )
            task = Task(
                name="some name", request=MagicMock(), locust_request=locust_request
            )
            expected_url = 'f"http://a.b.c/{some.value}/"'

            modified_task = Task.replace_url(task, expected_url)

            assert modified_task.locust_request.url == expected_url


class TestTask2:
    class TestFromTask:
        def test_without_locust_request_it_proxies_the_request(self):
            req = Mock(spec_set=Request)
            task = Task(name="T", request=req)
            task2 = Task2.from_task(task)

            assert task2.name == "T"
            assert task2.request == req
            assert len(task2.statements) == 1

            assign = task2.statements[0]
            assert isinstance(assign, py.Assignment)
            assert assign.lhs == "response"

            assert isinstance(assign.rhs, py.ExpressionView)
            assert assign.rhs.target() == task2.request
            assert assign.rhs.converter is req_to_expr

        def test_with_locust_request_it_proxies_it(self):
            lr = Mock(spec_set=LocustRequest)
            req = Mock(spec_set=Request)
            task = Task(name="T", request=req, locust_request=lr)
            task2 = Task2.from_task(task)

            assert task2.name == "T"
            assert task2.request == req
            assert len(task2.statements) == 1

            assign = task2.statements[0]
            assert isinstance(assign, py.Assignment)
            assert assign.lhs == "response"

            assert isinstance(assign.rhs, py.ExpressionView)
            assert assign.rhs.target() == lr
            assert assign.rhs.converter is lreq_to_expr


class _KindOfDict(enum.Flag):
    Text = enum.auto()
    Params = enum.auto()
    Both = Text | Params


_formats = sampled_from(("json", "www"))
_kinds_of_dicts = sampled_from(_KindOfDict)


# From http://www.softwareishard.com/blog/har-12-spec/#postData.
@composite
def har_post_dicts(draw, format=None):
    format = format or draw(_formats)
    if format == "json":
        d = {"mimeType": "application/json", "text": """{"a":"b", "c": "d"}"""}
        if draw(booleans()):
            d["params"] = []
        if draw(booleans()):
            d["comment"] = ""
        return d

    d = {"mimeType": "application/x-www-form-urlencoded"}
    kind = draw(_kinds_of_dicts)
    if kind & _KindOfDict.Text:
        d["text"] = "a=b&c=d"
        if draw(booleans()):
            d.setdefault("params", [])
    if kind & _KindOfDict.Params:
        d["params"] = [{"name": "a", "value": "b"}, {"name": "c", "value": "d"}]
        if draw(booleans()):
            d.setdefault("text", "")
    return d


class TestRequestPostData:
    def test_as_kwargs_only_shows_defined(self):
        v, w = MagicMock(), MagicMock()
        assert RequestsPostData(data=v).as_kwargs() == {"data": v}
        assert RequestsPostData(params=v, json=w).as_kwargs() == {
            "params": v,
            "json": w,
        }

    class TestFromHarPostData:
        @given(har_post_dicts(format="json"))
        def test_it_selects_json_approach_for_json_format(self, d: dict):
            rpd = RequestsPostData.from_har_post_data(d)
            assert rpd.json == py.Literal({"a": "b", "c": "d"})
            assert rpd.data is None

        @given(har_post_dicts(format="www"))
        def test_it_selects_data_approach_for_urlencoded_format(self, d: dict):
            rpd = RequestsPostData.from_har_post_data(d)
            assert rpd.json is None
            assert rpd.data == py.Literal(b"a=b&c=d") or rpd.params == py.Literal(
                [(b"a", b"b"), (b"c", b"d")]
            )

        @given(har_post_dicts())
        def test_it_doesnt_raise_error_on_valid_input(self, d: dict):
            RequestsPostData.from_har_post_data(d)

        def test_it_raises_on_post_data_without_text_or_params(self):
            with pytest.raises(ValueError):
                RequestsPostData.from_har_post_data({"mimeType": "nil"})

        def test_it_raises_on_invalid_json(self):
            with pytest.raises(ValueError):
                RequestsPostData.from_har_post_data(
                    {"mimeType": JSON_MIME_TYPE, "text": "not json"}
                )

        @pytest.mark.parametrize(
            "mime,kwarg,val",
            (
                (JSON_MIME_TYPE, "json", {}),
                ("application/x-www-form-urlencoded", "data", b"{}"),
            ),
        )
        def test_it_accepts_both_params_and_text(self, mime: str, kwarg, val):
            expected_fields = {
                "params": py.Literal([(b"n", b"v")]),
                kwarg: py.Literal(val),
            }
            assert RequestsPostData.from_har_post_data(
                {
                    "mimeType": mime,
                    "text": "{}",
                    "params": [{"name": "n", "value": "v"}],
                }
            ) == RequestsPostData(**expected_fields)


class TestReqToExpr:
    def test_it_supports_get_requests(self):
        url = "http://abc.de"
        r = Request(
            timestamp=MagicMock(),
            method=HttpMethod.GET,
            url=urlparse(url),
            headers=[Header("a", "b")],
            query=[QueryPair("x", "y")],  # query is currently ignored for GET
        )
        assert req_to_expr(r) == py.FunctionCall(
            name="self.client.get",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
            },
        )

    def test_it_supports_urlencoded_post_requests(self):
        url = "http://abc.de"
        r = Request(
            timestamp=MagicMock(),
            method=HttpMethod.POST,
            url=urlparse(url),
            headers=[Header("a", "b")],
            post_data={
                "mimeType": "application/x-www-form-urlencoded",
                "params": [{"name": "x", "value": "y"}],
                "text": "z=7",
            },
        )
        assert req_to_expr(r) == py.FunctionCall(
            name="self.client.post",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
                "data": py.Literal(b"z=7"),
                "params": py.Literal([(b"x", b"y")]),
            },
        )

    def test_it_supports_json_post_requests(self):
        url = "http://abc.de"
        r = Request(
            timestamp=MagicMock(),
            method=HttpMethod.POST,
            url=urlparse(url),
            headers=[Header("a", "b")],
            post_data={
                "mimeType": "application/json",
                "params": [{"name": "x", "value": "y"}],
                "text": """{"z": 7}""",
            },
        )
        assert req_to_expr(r) == py.FunctionCall(
            name="self.client.post",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
                "json": py.Literal({"z": 7}),
                "params": py.Literal([(b"x", b"y")]),
            },
        )

    def test_it_supports_empty_post_requests(self):
        url = "http://abc.de"
        r = Request(
            timestamp=MagicMock(),
            method=HttpMethod.POST,
            url=urlparse(url),
            headers=[Header("a", "b")],
            post_data=None,
        )
        assert req_to_expr(r) == py.FunctionCall(
            name="self.client.post",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
            },
        )

    def test_it_supports_put_requests_with_payload(self):
        url = "http://abc.de"
        r = Request(
            timestamp=MagicMock(),
            method=HttpMethod.PUT,
            url=urlparse(url),
            headers=[Header("a", "b")],
            query=[QueryPair("c", "d")],
            post_data={
                "mimeType": "application/json",
                "params": [{"name": "x", "value": "y"}],
                "text": """{"z": 7}""",
            },
        )
        assert req_to_expr(r) == py.FunctionCall(
            name="self.client.put",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
                "json": py.Literal({"z": 7}),
                "params": py.Literal([(b"x", b"y"), (b"c", b"d")]),
            },
        )

    def test_it_supports_put_requests_without_payload(self):
        url = "http://abc.de"
        r = Request(
            timestamp=MagicMock(),
            method=HttpMethod.PUT,
            url=urlparse(url),
            headers=[Header("a", "b")],
            query=[QueryPair("c", "d")],
            post_data=None,
        )
        assert req_to_expr(r) == py.FunctionCall(
            name="self.client.put",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
                "params": py.Literal([(b"c", b"d")]),
            },
        )

    def test_it_uses_the_custom_name_if_provided(self):
        url = "http://abc.de"
        name = "my-req"
        r = Request(
            name=name, timestamp=MagicMock(), method=HttpMethod.GET, url=urlparse(url)
        )
        assert req_to_expr(r) == py.FunctionCall(
            name="self.client.get",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(name),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
            },
        )


class TestLreqToExpr:
    def test_it_supports_get_requests(self):
        url = "http://abc.de"
        r = LocustRequest.from_request(
            Request(
                timestamp=MagicMock(),
                method=HttpMethod.GET,
                url=urlparse(url),
                headers=[Header("a", "b")],
                query=[QueryPair("x", "y")],  # query is currently ignored for GET
            )
        )
        assert lreq_to_expr(r) == py.FunctionCall(
            name="self.client.get",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
            },
        )

    def test_it_supports_fstring_urls(self):
        url = "http://abc.{tld}"
        r = LocustRequest(method=HttpMethod.GET, url=f"f'{url}'", headers={"a": "b"})
        assert lreq_to_expr(r) == py.FunctionCall(
            name="self.client.get",
            named_args={
                "url": py.FString(url),
                "name": py.FString(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
            },
        )

    def test_it_supports_urlencoded_post_requests(self):
        url = "http://abc.de"
        r = LocustRequest.from_request(
            Request(
                timestamp=MagicMock(),
                method=HttpMethod.POST,
                url=urlparse(url),
                headers=[Header("a", "b")],
                post_data={
                    "mimeType": "application/x-www-form-urlencoded",
                    "params": [{"name": "x", "value": "y"}],
                    "text": "z=7",
                },
            )
        )
        assert lreq_to_expr(r) == py.FunctionCall(
            name="self.client.post",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
                "data": py.Literal(b"z=7"),
                "params": py.Literal([(b"x", b"y")]),
            },
        )

    def test_it_supports_json_post_requests(self):
        url = "http://abc.de"
        r = LocustRequest.from_request(
            Request(
                timestamp=MagicMock(),
                method=HttpMethod.POST,
                url=urlparse(url),
                headers=[Header("a", "b")],
                post_data={
                    "mimeType": "application/json",
                    "params": [{"name": "x", "value": "y"}],
                    "text": """{"z": 7}""",
                },
            )
        )
        assert lreq_to_expr(r) == py.FunctionCall(
            name="self.client.post",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
                "json": py.Literal({"z": 7}),
                "params": py.Literal([(b"x", b"y")]),
            },
        )

    def test_it_supports_empty_post_requests(self):
        url = "http://abc.de"
        r = LocustRequest.from_request(
            Request(
                timestamp=MagicMock(),
                method=HttpMethod.POST,
                url=urlparse(url),
                headers=[Header("a", "b")],
                post_data=None,
            )
        )
        assert lreq_to_expr(r) == py.FunctionCall(
            name="self.client.post",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
            },
        )

    def test_it_supports_put_requests_with_payload(self):
        url = "http://abc.de"
        r = LocustRequest.from_request(
            Request(
                timestamp=MagicMock(),
                method=HttpMethod.PUT,
                url=urlparse(url),
                headers=[Header("a", "b")],
                query=[QueryPair("c", "d")],
                post_data={
                    "mimeType": "application/json",
                    "params": [{"name": "x", "value": "y"}],
                    "text": """{"z": 7}""",
                },
            )
        )
        assert lreq_to_expr(r) == py.FunctionCall(
            name="self.client.put",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
                "json": py.Literal({"z": 7}),
                "params": py.Literal([(b"x", b"y"), (b"c", b"d")]),
            },
        )

    def test_it_supports_put_requests_without_payload(self):
        url = "http://abc.de"
        r = LocustRequest.from_request(
            Request(
                timestamp=MagicMock(),
                method=HttpMethod.PUT,
                url=urlparse(url),
                headers=[Header("a", "b")],
                query=[QueryPair("c", "d")],
                post_data=None,
            )
        )
        assert lreq_to_expr(r) == py.FunctionCall(
            name="self.client.put",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(url),
                "headers": py.Literal({"a": "b"}),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
                "params": py.Literal([(b"c", b"d")]),
            },
        )

    def test_it_uses_the_custom_name_if_provided(self):
        url = "http://abc.de"
        name = "my-req"
        r = LocustRequest.from_request(
            Request(
                name=name,
                timestamp=MagicMock(),
                method=HttpMethod.GET,
                url=urlparse(url),
            )
        )
        assert lreq_to_expr(r) == py.FunctionCall(
            name="self.client.get",
            named_args={
                "url": py.Literal(url),
                "name": py.Literal(name),
                "timeout": py.Literal(TIMEOUT),
                "allow_redirects": py.Literal(False),
            },
        )



import logging
import os
from pathlib import Path

import transformer
from transformer.helpers import DUMMY_HAR_STRING


def test_dummy_plugin_works(tmp_path: Path, caplog):
    har_path = tmp_path / "test.har"
    har_path.write_text(DUMMY_HAR_STRING)

    caplog.set_level(logging.INFO)

    with open(os.path.devnull, "w") as f:
        transformer.dump(f, [har_path], plugins=["transformer.plugins.dummy"])

    assert "The first request was https://www.zalando.de" in caplog.text
